/*---------------------------------------------*/
/*	GDP15 - 2014/15
 *
 * 	Author:		Stuart Witts
 * 	Email:		stw1g11@ecs.soton.ac.uk
 * 	Date:		January 2014
 *
 *  ---------------------------
 * 	Base Unit embedded software
 * 	---------------------------
 * 		- UART communication via BlueTooth
 * 			- Configuration
 * 			- Data retrieval
 * 		- SPI communication to sensor modules
 * 			- Identification
 * 			- Sampling
 * 		- SD card logging - FAT32 over SPI
 * 			- Sampled data
 * 		- RTC (Real-Time Clock)
 *
 * */
/*---------------------------------------------*/


/*---------------------------------------------*/
/* Includes */
/*---------------------------------------------*/

#include "main.h"

extern void disk_timerproc(void);


/*---------------------------------------------*/
/* Global Variables */
/*---------------------------------------------*/

/* Interfaces */
RCC_ClocksTypeDef RCC_Clocks;
EXTI_InitTypeDef EXTI_InitStructure;
NVIC_InitTypeDef NVIC_InitStructure;
GPIO_InitTypeDef GPIO_InitStruct;
USART_InitTypeDef USART_InitStruct;
SPI_InitTypeDef SPI_InitStruct;
I2C_InitTypeDef I2C_InitStruct;

/* USART */
static int tx_index_usart2 = 0;
static int rx_index_usart2 = 0;
static int bt_configured = 0;
static int bt_receiving = 0;
static int bt_received = 0;
static int bt_live = 0;
uint8_t bt_frame[2];
uint8_t buffer_usart2[MAX_CHARS+1]; // MAX_CHARS + '\r' or similar termination character
#ifdef DEBUG_USE_USART1
	static int tx_index_usart1 = 0;
	static int rx_index_usart1 = 0;
	uint8_t buffer_usart1[MAX_CHARS+1]; // MAX_CHARS + '\r' or similar termination character
#endif

/* RTC */
uint8_t buffer_time[9];		// HH:MM:SS
uint8_t* buffer_day;		// Sunday (1) --> Saturday (7)
uint8_t buffer_date[11];	// DD-MM-YYYY
const uint8_t time_separator = ':';
const uint8_t date_separator = '-';
// Day Of Week (dow) storage
uint8_t dow[7][10] = {
		"Sunday",
		"Monday",
		"Tuesday",
		"Wednesday",
		"Thursday",
		"Friday",
		"Saturday"
};

/* SD */
int sd_present = 0;				// SD_PRESENT (see EXTI0_IRQHandler())
int sd_locked = 0;				// SD_/WP (see EXTI0_IRQHandler())
int sd_isr_mount = 0;			// monitored in main() which calls ConfigureSD() if 1
int sd_isr_mount_open = 0;		// set in ConfigureSD() monitored in main() which calls SD_OpenLog() if 1
int sd_isr_mount_config = 0;	// set in ConfigureSD() monitored in main() which calls SD_CheckConfig() if 1
int sd_isr_unmount = 0;			// monitored in main() which calls SD_Unmount() if 1
int sd_mounted = 0;				// 1=SD card mounted, 0=SD card not mounted
FATFS fs;						// File System object structure
FILINFO fno;					// File Information structure
FIL fp;							// File object structure
FIL fpConfig;					// File object structure (config)
char frbuff[33];				// file function return code (string)
char configFilename[13];		// filename {8}+'.'+{3}+'\0' (13 chars)
char logFilename[13];			// filename {8}+'.'+{3}+'\0' (13 chars)
char rbuff[MAX_CHARS+1];		// buffer (read)
char wbuff[MAX_CHARS+1];		// buffer (write)

/* Sensors */
uint8_t sensorType[4] = {	// 'P'=80=Photovoltaic, 'T'=84=Thermoelectric
		0, 0, 0, 0
};
uint8_t sensorData[4][SENSOR_MAX_BYTES+1];	// data storage for each sensor module
char sensorLogData[MAX_CHARS+1];			// buffer for converting sensorData into format which can be logged
uint32_t joinData = 0;						// used to recombine 8-bit packets
uint32_t TimeOutSEN = 0;					// TIMEOUT used for SEN_GetData()

/* Logging */
int log_isr_start = 0;			// monitored in main() which calls Logging_Start() if 1
int log_isr_stop = 0;			// monitored in main() which calls Logging_Stop() if 1
int log_run = 0;				// logging? 1=started, 0=stopped
uint32_t logPeriod = 1000;		// log period; default = 1000ms
uint32_t logTimer = 0;			// used to keep log period


/*---------------------------------------------*/
/* Main */
/*---------------------------------------------*/

int main(void)
{
	/* SysTick end of count event each 1ms */
	RCC_GetClocksFreq(&RCC_Clocks);
	SysTick_Config(RCC_Clocks.HCLK_Frequency / 1000);

	EnableClocks();
	ConfigureEXTI();
	ConfigureNVIC();
	ConfigureGPIO();
	#ifdef DEBUG_USE_USART1
		ConfigureUSART1();
	#endif
	ConfigureUSART2();
	ConfigureSPI2();
	ConfigureI2C2();
	ConfigureRTC();
	ConfigureSD();
	ConfigureSensors();
	ConfigureBluetooth();

	SD_CheckConfig();

	while (1)
	{
		/* Handle SD card insert/remove */
		if (sd_isr_mount)
			ConfigureSD();
		if (sd_isr_mount_config)
			SD_CheckConfig();
		if (sd_isr_mount_open)
			SD_OpenLog();
		if (sd_isr_unmount)
			SD_Unmount();

		/* Handle BT data received */
		if (bt_received)
			BT_Check();

		/* Handle logging start/stop */
		if (log_isr_start)
			Logging_Start();
		if (log_isr_stop)
			Logging_Stop();

		// DEBUG force enable live readings
//		bt_live = 1;

		// DEBUG simulate start and stop logging
//		if (mainCount == 1)
//			log_isr_start = 1;		// simulate start
//		if (mainCount == 10)
//			log_isr_stop = 1;		// simulate stop


		/* Main Process */

		logTimer = logPeriod;	// start log timer

		SEN_GetData();			// get data from sensor modules
		SEN_FormatData();		// recombine packets of data and format ready to be logged

		// Clear write buffer (wbuff)
		wbuff[0] = 0;
		// Renew TIMESTAMP
		I2C_RTC_Buffer();
		// Fill write buffer (wbuff)
		sprintf(wbuff, "%s.%s,%s", buffer_date, buffer_time, sensorLogData);

		if (sd_mounted && log_run)
		{
			SD_LogWriteLine();

			// Flash LED_EXT
			PinSet(GPIOB, GPIO_Pin_12);
			Delay(10);
			PinReset(GPIOB, GPIO_Pin_12);
		}

		if (bt_live)
		{
			if (sd_mounted)
				SendChar_USART2((uint8_t)'M');
			else
				SendChar_USART2((uint8_t)'U');
			if (log_run)
				SendChar_USART2((uint8_t)'Y');
			else
				SendChar_USART2((uint8_t)'N');

			SendString_USART2((uint8_t*)wbuff);
		}

		// DEBUG use for simulated start
//		mainCount++;

		//DEBUG provide LFCR for console output
//		SendString_USART1((uint8_t*)"\n\r");

		// Turn off LED_EXT
		PinReset(GPIOB, GPIO_Pin_12);

		while (logTimer != 0);	// wait until end of log period
	}
}


/*---------------------------------------------*/
/* ISRs */
/*---------------------------------------------*/

void EXTI0_IRQHandler()
{
	if (EXTI_GetITStatus(EXTI_Line0) == SET)
	{
		/* Clear EXTI line 0 pending bit */
		EXTI_ClearITPendingBit(EXTI_Line0);

		uint16_t portc = PortIDR(GPIOC);
		if (portc & GPIO_Pin_0)
		{
			/* SD card present (PC0) */

			if (sd_present)
				return;	// SD card already present (erroneous ISR fire)

			sd_present = 1;

			#ifdef DEBUG_USE_USART1
				SendString_USART1((uint8_t*)"SD card inserted.\n\r");
			#endif
			if (portc & GPIO_Pin_1)
			{
				#ifdef DEBUG_USE_USART1
					SendString_USART1((uint8_t*)"  Unlocked.\n\r");
				#endif
				sd_locked = 0;

				sd_isr_mount = 1;	// Value monitored in main() which calls ConfigureSD() if 1
			}
			else
			{
				#ifdef DEBUG_USE_USART1
					SendString_USART1((uint8_t*)"  Locked.\n\r");
				#endif
				sd_locked = 1;
			}
		}
		else
		{
			/* No SD card present (!PC0) */

			if (!sd_present)
				return;	// SD card already removed (erroneous ISR fire)

			sd_present = 0;

			#ifdef DEBUG_USE_USART1
				SendString_USART1((uint8_t*)"SD card removed!\n\r");
			#endif

			sd_isr_unmount = 1;	// Value monitored in main() which calls SD_Unmount() if 1

		}
	}
}

#ifdef DEBUG_USE_USART1
void USART1_IRQHandler()
{
	if (USART_GetITStatus(USART1, USART_IT_TXE) != RESET) // Transmit the string in a loop
	{
		USART_SendData(USART1, buffer_usart1[tx_index_usart1++]);

		if (tx_index_usart1 >= MAX_CHARS)
			tx_index_usart1 = 0;
	}

	if (USART_GetITStatus(USART1, USART_IT_RXNE) != RESET) // Received characters modify string
	{
		buffer_usart1[rx_index_usart1++] = USART_ReceiveData(USART1);

		// Handle "Start Logging" and "Stop Logging" commands to make debugging easier from console
		if (buffer_usart1[rx_index_usart1 - 1] == 'Y')
			log_isr_start = 1;
		if (buffer_usart1[rx_index_usart1 - 1] == 'N')
			log_isr_stop = 1;

		if (rx_index_usart1 >= MAX_CHARS)
			rx_index_usart1 = 0;
	}
}
#endif

void USART2_IRQHandler()
{
	if (USART_GetITStatus(USART2, USART_IT_TXE) != RESET) // Transmit the string in a loop
	{
		USART_SendData(USART2, buffer_usart2[tx_index_usart2++]);

		if (tx_index_usart2 >= MAX_CHARS)
			tx_index_usart2 = 0;
	}

	if (USART_GetITStatus(USART2, USART_IT_RXNE) != RESET) // Received characters modify string
	{
		buffer_usart2[rx_index_usart2++] = USART_ReceiveData(USART2);

		if (bt_configured)
		{
			if (!bt_receiving)
			{
				bt_frame[0] = bt_frame[1];
				bt_frame[1] = buffer_usart2[rx_index_usart2 - 1];

				if (bt_frame[0] == 'S' && bt_frame[1] == 'T')
				{
					rx_index_usart2 = 0;
					bt_receiving = 1;
				}
			}
			else
			{
				bt_frame[0] = bt_frame[1];
				bt_frame[1] = buffer_usart2[rx_index_usart2 - 1];

				if (bt_frame[0] == 'T' && bt_frame[1] == 'S')
				{
					bt_receiving = 0;
					buffer_usart2[rx_index_usart2 - 2] = 0;
					bt_received = 1;
				}
			}
		}

		if (rx_index_usart2 >= MAX_CHARS)
			rx_index_usart2 = 0;
	}
}


/*---------------------------------------------*/
/* Functions */
/*---------------------------------------------*/

void EnableClocks()
{
	RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOA, ENABLE);
	RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOB, ENABLE);
	RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOC, ENABLE);

	RCC_APB2PeriphClockCmd(RCC_APB2Periph_SYSCFG, ENABLE);	// EXTILineConfig()

	RCC_APB1PeriphClockCmd(RCC_APB1Periph_USART2, ENABLE);
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_SPI2, ENABLE);
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_SPI3, ENABLE);
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_I2C2, ENABLE);

	#ifdef DEBUG_USE_USART1
		// Using PA9 (USART1_TX), PA10 (USART1_RX)
		RCC_APB2PeriphClockCmd(RCC_APB2Periph_USART1, ENABLE);
	#endif

	return;
}

/******************************************************************************/

void ConfigureEXTI()
{
	EXTI_DeInit();

	/* Connect EXTI0 line to SD_PRESENT (PC0) */
	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOC, EXTI_PinSource0);

	/* Configure EXTI0 line */
	EXTI_InitStructure.EXTI_Line	= EXTI_Line0;
	EXTI_InitStructure.EXTI_Mode	= EXTI_Mode_Interrupt;
	EXTI_InitStructure.EXTI_Trigger	= EXTI_Trigger_Rising_Falling;
	EXTI_InitStructure.EXTI_LineCmd	= ENABLE;

	/* Initialise EXTI0 line */
	EXTI_Init(&EXTI_InitStructure);
}

/******************************************************************************/

void ConfigureNVIC()
{
	/* NVIC_PriorityGroup_0 does not allow nested interrupts */
	NVIC_PriorityGroupConfig(NVIC_PriorityGroup_1);

	/* Enable EXTI0 IRQ Channel */
	NVIC_InitStructure.NVIC_IRQChannel = EXTI0_IRQn;
	NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 1;
	NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0;
	NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
	NVIC_Init(&NVIC_InitStructure);

	/* Enable USART2 IRQ Channel */
	NVIC_InitStructure.NVIC_IRQChannel = USART2_IRQn;
	NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 0;
	NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0;
	NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
	NVIC_Init(&NVIC_InitStructure);
	
	#ifdef DEBUG_USE_USART1
		/* Enable USART1 IRQ Channel */
		NVIC_InitStructure.NVIC_IRQChannel = USART1_IRQn;
		NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 0;
		NVIC_InitStructure.NVIC_IRQChannelSubPriority = 1;
		NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
		NVIC_Init(&NVIC_InitStructure);
	#endif
}

/******************************************************************************/

void ConfigureGPIO()
{
	/*******************************************/
	/* Connect pins to Alternate Function (AF) */
	/*******************************************/
	/* USART2 - BlueTooth Module */
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource0, GPIO_AF_USART2);	// USART2_CTS (PA0)
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource1, GPIO_AF_USART2);	// USART2_RTS (PA1)
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource2, GPIO_AF_USART2);	// USART2_TX (PA2)
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource3, GPIO_AF_USART2);	// USART2_RX (PA3)
	#ifdef DEBUG_USE_USART1
		/*USART1 - DEBUG Output */
		GPIO_PinAFConfig(GPIOA, GPIO_PinSource9, GPIO_AF_USART1);	// USART1_TX (PA9)
		GPIO_PinAFConfig(GPIOA, GPIO_PinSource10, GPIO_AF_USART1);	// USART1_RX (PA10)
	#endif

	/* SPI2 - Sensors */
//	GPIO_PinAFConfig(GPIOB, GPIO_PinSource12, GPIO_AF_SPI2);	// SPI2_NSS (PB12) // DEBUG use for debugging SPI2
	GPIO_PinAFConfig(GPIOB, GPIO_PinSource13, GPIO_AF_SPI2);	// SPI2_SCK (PB13)
	GPIO_PinAFConfig(GPIOB, GPIO_PinSource14, GPIO_AF_SPI2);	// SPI2_MISO (PB14)
	GPIO_PinAFConfig(GPIOB, GPIO_PinSource15, GPIO_AF_SPI2);	// SPI2_MOSI (PB15)

	/* SPI3 - SD Card */
	GPIO_PinAFConfig(GPIOC, GPIO_PinSource10, GPIO_AF_SPI3);	// SPI3_SCK (PC10)
	GPIO_PinAFConfig(GPIOC, GPIO_PinSource11, GPIO_AF_SPI3);	// SPI3_MISO (PC11)
	GPIO_PinAFConfig(GPIOC, GPIO_PinSource12, GPIO_AF_SPI3);	// SPI3_MOSI (PC12)

	/* I2C2 - RTC */
	GPIO_PinAFConfig(GPIOB, GPIO_PinSource10, GPIO_AF_I2C2);	// I2C2_SCL (PB10)
	GPIO_PinAFConfig(GPIOB, GPIO_PinSource11, GPIO_AF_I2C2);	// I2C2_SDA (PB11)


	/*********************************************/
	/* Configure GPIO AF with Push Pull, No Pull */
	/*********************************************/
	GPIO_InitStruct.GPIO_Mode	= GPIO_Mode_AF;
	GPIO_InitStruct.GPIO_OType	= GPIO_OType_PP;
	GPIO_InitStruct.GPIO_PuPd	= GPIO_PuPd_NOPULL;
	GPIO_InitStruct.GPIO_Speed	= GPIO_Speed_40MHz;

	/* Initialise USART2 GPIO AF - BlueTooth Module */
	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_0;
	GPIO_Init(GPIOA, &GPIO_InitStruct);	// USART2_CTS (PA0)

	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_1;
	GPIO_Init(GPIOA, &GPIO_InitStruct);	// USART2_RTS (PA1)

	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_2;
	GPIO_Init(GPIOA, &GPIO_InitStruct);	// USART2_TX (PA2)
	
	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_3;
	GPIO_Init(GPIOA, &GPIO_InitStruct);	// USART2_RX (PA3)

	#ifdef DEBUG_USE_USART1
		/* Initialise USART1 GPIO AF - DEBUG Output */
		GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_9;
		GPIO_Init(GPIOA, &GPIO_InitStruct);	// USART1_TX (PA9)
		
		GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_10;
		GPIO_Init(GPIOA, &GPIO_InitStruct);	// USART1_RX (PA10)
	#endif
	
	/* Initialise SPI2 GPIO AF - Sensors  */
//	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_12;				//DEBUG use for debugging SPI2
//	GPIO_Init(GPIOB, &GPIO_InitStruct);	// SPI2_NSS (PB12)	//DEBUG use for debugging SPI2

	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_13;
	GPIO_Init(GPIOB, &GPIO_InitStruct);	// SPI2_SCK (PB13)

	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_14;
	GPIO_Init(GPIOB, &GPIO_InitStruct);	// SPI2_MISO (PB14)

	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_15;
	GPIO_Init(GPIOB, &GPIO_InitStruct);	// SPI2_MOSI (PB15)

	/* Initialise SPI3 GPIO AF - SD Card */
	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_10;
	GPIO_Init(GPIOC, &GPIO_InitStruct);	// SPI3_SCK (PC10)

	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_11;
	GPIO_Init(GPIOC, &GPIO_InitStruct);	// SPI3_MISO (PC11)

	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_12;
	GPIO_Init(GPIOC, &GPIO_InitStruct);	// SPI3_MOSI (PC12)


	/**********************************************/
	/* Configure GPIO AF with Open Drain, No Pull */
	/**********************************************/
	GPIO_InitStruct.GPIO_Mode	= GPIO_Mode_AF;
	GPIO_InitStruct.GPIO_OType	= GPIO_OType_OD;
	GPIO_InitStruct.GPIO_PuPd	= GPIO_PuPd_NOPULL;
	GPIO_InitStruct.GPIO_Speed	= GPIO_Speed_40MHz;

	/* Initialise I2C2 GPIO AF - RTC */
	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_10;
	GPIO_Init(GPIOB, &GPIO_InitStruct);	// I2C2_SCL (PB10)

	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_11;
	GPIO_Init(GPIOB, &GPIO_InitStruct);	// I2C2_SDA (PB11)
	

	/**********************************/
	/* Configure GPIO IN with No Pull */
	/**********************************/
	GPIO_InitStruct.GPIO_Mode	= GPIO_Mode_IN;
	GPIO_InitStruct.GPIO_OType	= GPIO_OType_PP;
	GPIO_InitStruct.GPIO_PuPd	= GPIO_PuPd_NOPULL;
	GPIO_InitStruct.GPIO_Speed	= GPIO_Speed_40MHz;

	/* Initialise SD_PRESENT (PC0) */
	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_0;
	GPIO_Init(GPIOC, &GPIO_InitStruct); // SD_PRESENT (PC0) - SD_CD (Card Detect)

	/* Initialise SD_/WP (PC1) */
	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_1;
	GPIO_Init(GPIOC, &GPIO_InitStruct); // SD_/WP (PC1) - (Write Protect - active low)


	/***********************************/
	/* Configure GPIO OUT with No Pull */
	/***********************************/
	GPIO_InitStruct.GPIO_Mode	= GPIO_Mode_OUT;
	GPIO_InitStruct.GPIO_OType	= GPIO_OType_PP;
	GPIO_InitStruct.GPIO_PuPd	= GPIO_PuPd_NOPULL;
	GPIO_InitStruct.GPIO_Speed	= GPIO_Speed_40MHz;

	/* Initialise SD_SPI_SS (PB4) - SD_CS */
	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_4;
	GPIO_Init(GPIOB, &GPIO_InitStruct);	// SD_SPI_SS (PB4) - SD_CS


	/***********************************/
	/* Configure GPIO OUT with Pull Up */
	/***********************************/
	GPIO_InitStruct.GPIO_Mode	= GPIO_Mode_OUT;
	GPIO_InitStruct.GPIO_OType	= GPIO_OType_PP;
	GPIO_InitStruct.GPIO_PuPd	= GPIO_PuPd_UP;
	GPIO_InitStruct.GPIO_Speed	= GPIO_Speed_40MHz;

	/* Initialise SEN_SPI_SS (PA8) - SEN_3_SS */
	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_8;
	GPIO_Init(GPIOA, &GPIO_InitStruct);	// SEN_3_SS (PA8)
#ifndef DEBUG_USE_USART1
	/* Initialise SEN_SPI_SS (PA9) - SEN_2_SS */
	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_9;
	GPIO_Init(GPIOA, &GPIO_InitStruct);	// SEN_2_SS (PA9)

	/* Initialise SEN_SPI_SS (PA10) - SEN_4_SS */
	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_10;
	GPIO_Init(GPIOA, &GPIO_InitStruct);	// SEN_4_SS (PA10)
#endif
	/* Initialise SEN_SPI_SS (PA11) - SEN_1_SS */
	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_11;
	GPIO_Init(GPIOA, &GPIO_InitStruct);	// SEN_1_SS (PA11)


	/*************************************/
	/* Configure GPIO OUT with Pull Down */
	/*************************************/
	GPIO_InitStruct.GPIO_Mode	= GPIO_Mode_OUT;
	GPIO_InitStruct.GPIO_OType	= GPIO_OType_PP;
	GPIO_InitStruct.GPIO_PuPd	= GPIO_PuPd_DOWN;
	GPIO_InitStruct.GPIO_Speed	= GPIO_Speed_40MHz;

	/* Initialise LED_EXT (PB12) */
	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_12;
	GPIO_Init(GPIOB, &GPIO_InitStruct); // LED_EXT (PB12)

	/* Initialise BT_CMD (PC2) */
	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_2;
	GPIO_Init(GPIOC, &GPIO_InitStruct); // BT_CMD (PC2)

	/* Initialise BT_WK_SW (PC3) */
	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_3;
	GPIO_Init(GPIOC, &GPIO_InitStruct); // BT_WK_SW (PC3)

	/* Initialise BT_WAKE_HW (PC14) */
	GPIO_InitStruct.GPIO_Pin	= GPIO_Pin_14;
	GPIO_Init(GPIOC, &GPIO_InitStruct); // BT_WAKE_HW (PC14)

	return;
}

/******************************************************************************/

#ifdef DEBUG_USE_USART1
void ConfigureUSART1()
{
	/* Before using this function:
	 * 1) EnableClocks()
	 * 2) ConfigureGPIO()
	 * */

	/* Reset USART1 interface */
	USART_DeInit(USART1);

	/* Configure USART1 interface */
	USART_InitStruct.USART_Mode					= USART_Mode_Tx | USART_Mode_Rx;
	USART_InitStruct.USART_BaudRate				= 115200;
	USART_InitStruct.USART_WordLength			= USART_WordLength_8b;
	USART_InitStruct.USART_StopBits				= USART_StopBits_1;
	USART_InitStruct.USART_Parity				= USART_Parity_No;
	USART_InitStruct.USART_HardwareFlowControl	= USART_HardwareFlowControl_None;

	/* Initialise USART1 interface */
	USART_Init(USART1, &USART_InitStruct);

	/* Enable USART1 interface */
	USART_Cmd(USART1, ENABLE);

	/* Enable USART1 interrupts RXNE and TXE */
	USART_ITConfig(USART1, USART_IT_RXNE, ENABLE);
	//USART_ITConfig(USART1, USART_IT_TXE, ENABLE);

	return;
}
#endif


void ConfigureUSART2()
{
	/* Before using this function:
	 * 1) EnableClocks()
	 * 2) ConfigureGPIO()
	 * */

	/* Reset USART2 interface */
	USART_DeInit(USART2);

	/* Configure USART2 interface */
	USART_InitStruct.USART_Mode					= USART_Mode_Tx | USART_Mode_Rx;
	USART_InitStruct.USART_BaudRate				= 115200;
	USART_InitStruct.USART_WordLength			= USART_WordLength_8b;
	USART_InitStruct.USART_StopBits				= USART_StopBits_1;
	USART_InitStruct.USART_Parity				= USART_Parity_No;
	USART_InitStruct.USART_HardwareFlowControl	= USART_HardwareFlowControl_None;

	/* Initialise USART2 interface */
	USART_Init(USART2, &USART_InitStruct);

	/* Enable USART2 interface */
	USART_Cmd(USART2, ENABLE);

	/* Enable USART2 interrupts RXNE and TXE */
	USART_ITConfig(USART2, USART_IT_RXNE, ENABLE);
	//USART_ITConfig(USART2, USART_IT_TXE, ENABLE);

	return;
}


void ConfigureSPI2()
{
	/* Before using this function:
	 * 1) EnableClocks()
	 * 2) ConfigureGPIO()
	 * */

	/* Reset SPI2 interface */
	SPI_I2S_DeInit(SPI2);

	/* Common SPI configuration */
	SPI_InitStruct.SPI_BaudRatePrescaler	= SPI_BaudRatePrescaler_128;
	SPI_InitStruct.SPI_CPHA					= SPI_CPHA_1Edge;
	SPI_InitStruct.SPI_CPOL					= SPI_CPOL_Low;
	SPI_InitStruct.SPI_DataSize				= SPI_DataSize_8b;
	SPI_InitStruct.SPI_Direction			= SPI_Direction_2Lines_FullDuplex;
	SPI_InitStruct.SPI_FirstBit				= SPI_FirstBit_MSB;
	SPI_InitStruct.SPI_NSS					= SPI_NSS_Soft;

	/* Initialise SPI2 interface */
	#ifdef SPI2_MASTER
		SPI_InitStruct.SPI_Mode					= SPI_Mode_Master;
	#else
		SPI_InitStruct.SPI_Mode					= SPI_Mode_Slave;
	#endif
	SPI_Init(SPI2, &SPI_InitStruct);

	/* Enable SPI_I2S_IT_RXNE interrupt for SPI2 */
	//SPI_I2S_ITConfig(SPI2, SPI_I2S_IT_RXNE, ENABLE);

	/* Enable SPI2 interface */
	SPI_Cmd(SPI2, ENABLE);

	return;
}


void ConfigureSPI3(uint16_t SPI_BaudRatePrescaler)
{
	/* Before using this function:
	 * 1) EnableClocks()
	 * 2) ConfigureGPIO()
	 * */

	/* Reset SPI3 interface */
	SPI_I2S_DeInit(SPI3);

	/* Common SPI configuration */
	SPI_InitStruct.SPI_BaudRatePrescaler	= SPI_BaudRatePrescaler;	// function argument
	SPI_InitStruct.SPI_CPHA					= SPI_CPHA_1Edge;
	SPI_InitStruct.SPI_CPOL					= SPI_CPOL_Low;
	SPI_InitStruct.SPI_DataSize				= SPI_DataSize_8b;
	SPI_InitStruct.SPI_Direction			= SPI_Direction_2Lines_FullDuplex;
	SPI_InitStruct.SPI_FirstBit				= SPI_FirstBit_MSB;
	SPI_InitStruct.SPI_NSS					= SPI_NSS_Soft;

	/* Initialise SPI3 interface */
	#ifdef SPI3_MASTER
		SPI_InitStruct.SPI_Mode					= SPI_Mode_Master;
	#else
		SPI_InitStruct.SPI_Mode					= SPI_Mode_Slave;
	#endif
	SPI_Init(SPI3, &SPI_InitStruct);

	/* Enable SPI_I2S_IT_RXNE interrupt for SPI3 */
	//SPI_I2S_ITConfig(SPI3, SPI_I2S_IT_RXNE, ENABLE);

	/* Enable SPI3 interface */
	SPI_Cmd(SPI3, ENABLE);

	return;
}


void ConfigureI2C2()
{
	/* Before using this function:
	 * 1) EnableClocks()
	 * 2) ConfigureGPIO()
	 * */

	/* Reset I2C2 interface */
	I2C_DeInit(I2C2);

	/* Configure I2C2 interface */
	I2C_InitStruct.I2C_Mode					= I2C_Mode_I2C;
	I2C_InitStruct.I2C_Ack					= I2C_Ack_Disable;
	I2C_InitStruct.I2C_AcknowledgedAddress	= I2C_AcknowledgedAddress_7bit;
	I2C_InitStruct.I2C_ClockSpeed			= 100000;	// 100kHz (max. 400kHz)
	I2C_InitStruct.I2C_DutyCycle			= I2C_DutyCycle_2;
	I2C_InitStruct.I2C_OwnAddress1			= 0x10;	// Not relevant in master mode

	/* Initialise I2C2 interface */
	I2C_Init(I2C2, &I2C_InitStruct);

	/* Enable I2C2 interface */
	I2C_Cmd(I2C2, ENABLE);

	return;
}

/******************************************************************************/

int ConfigureRTC()
{
	uint8_t value = 0; // Used for validation

	#ifdef DEBUG_USE_USART1
		SendString_USART1((uint8_t*)"\r\n-----------------------\r\n   RTC Configuration   \r\n-----------------------\r\n");
	#endif

	/* Write 0x00 to CFG1, TCH2, CFG2 */
	#ifdef DEBUG_USE_USART1
		SendString_USART1((uint8_t*)"Write 0x00 to CFG1, TCH2, CFG2.\r\n");
	#endif
	SendChar_I2C2(RTC_ADDR_CFG1, 0x00);
	SendChar_I2C2(RTC_ADDR_TCH2, 0x00);
	SendChar_I2C2(RTC_ADDR_CFG2, 0x00);

	/* Check RTC has started */
	#ifdef DEBUG_USE_USART1
		SendString_USART1((uint8_t*)"Check RTC has started.\r\n");
	#endif
	I2C_RTC_Start();

	/* Validate register values */
	#ifdef DEBUG_USE_USART1
		SendString_USART1((uint8_t*)"Validate register values...\r\n");
	#endif
	value = I2C_RTC_GetSeconds();
	if ((value<0) && (value>59))
	{
		#ifdef DEBUG_USE_USART1
			SendString_USART1((uint8_t*)"Reset SECONDS.\r\n");
		#endif
		I2C_RTC_SetSeconds(0);
	}
	value = I2C_RTC_GetMinutes();
	if ((value<0) && (value>59))
	{
		#ifdef DEBUG_USE_USART1
			SendString_USART1((uint8_t*)"Reset MINUTES.\r\n");
		#endif
		I2C_RTC_SetMinutes(0);
	}
	value = I2C_RTC_GetHours();
	if ((value<0) && (value>23))
	{
		#ifdef DEBUG_USE_USART1
			SendString_USART1((uint8_t*)"Reset HOURS.\r\n");
		#endif
		I2C_RTC_SetHours(0);
	}
	value = I2C_RTC_GetDay();
	if ((value<1) && (value>7))
	{
		#ifdef DEBUG_USE_USART1
			SendString_USART1((uint8_t*)"Reset DAY.\r\n");
		#endif
		I2C_RTC_SetDay(1);
	}
	value = I2C_RTC_GetDate();
	if ((value<0) && (value>31))
	{
		#ifdef DEBUG_USE_USART1
			SendString_USART1((uint8_t*)"Reset DATE.\r\n");
		#endif
		I2C_RTC_SetDate(0);
	}
	value = I2C_RTC_GetMonth();
	if ((value<0) && (value>12))
	{
		#ifdef DEBUG_USE_USART1
			SendString_USART1((uint8_t*)"Reset MONTH.\r\n");
		#endif
		I2C_RTC_SetMonth(0);
	}
	value = I2C_RTC_GetYear();
	if ((value<0) && (value>99))
	{
		#ifdef DEBUG_USE_USART1
			SendString_USART1((uint8_t*)"Reset YEARS.\r\n");
		#endif
		I2C_RTC_SetYear(0);
	}
	#ifdef DEBUG_USE_USART1
		SendString_USART1((uint8_t*)"...Done.\r\n");
	#endif

	return 0;
}
int ConfigureSD()
{
	#ifdef DEBUG_USE_USART1
		SendString_USART1((uint8_t*)"\r\n-----------------------\r\n    SD Configuration   \r\n-----------------------\r\n");
	#endif

	uint16_t portc = PortIDR(GPIOC);
	if (portc & GPIO_Pin_0)
	{
		/* SD card present (PC0) */
		sd_present = 1;

		if (portc & GPIO_Pin_1)
		{
			sd_locked = 0;
		}
		else
		{
			sd_locked = 1;
			#ifdef DEBUG_USE_USART1
				SendString_USART1((uint8_t*)"Remove SD card lock!\n\r");
			#endif
			return 2;
		}
	}
	else
	{
		/* No SD card present (!PC0) */
		sd_present = 0;
		#ifdef DEBUG_USE_USART1
			SendString_USART1((uint8_t*)"Insert SD card!\n\r");
		#endif
		return 1;
	}

	int res = SD_Mount();

	if (!res)
		sd_mounted = 1;
	else
		sd_mounted = 0;


	if (sd_mounted && sd_isr_mount)
	{
		// if SD card inserted, trigger config check (monitored in main())
//		SendString_USART1((uint8_t*)"\n\rsd_isr_mount_config = 1\n\r"); //DEBUG check code not dead
		sd_isr_mount_config = 1;
	}

	if (sd_mounted && log_run)
	{
		// if SD card inserted while log_run, trigger file open (monitored in main())
//		SendString_USART1((uint8_t*)"\n\rsd_isr_mount_open = 1\n\r");	//DEBUG check code not dead
		sd_isr_mount_open = 1;
	}

	sd_isr_mount = 0;	// reset flag in case ConfigureSD() was triggered by ISR

	return res;
}
int ConfigureSensors()
{
	uint8_t startBytes[2];
	uint8_t stopBytes[2];

	int i, b;
	for (i = 0; i < 4; i++)
	{
		startBytes[0] = 0;
		startBytes[1] = 0;
		stopBytes[0] = 0;
		stopBytes[1] = 0;

		// Select Sensor i
		SEN_Select(i);

		// Request data
		SendChar_SPI2('I');

		// Wait for start bytes
		TimeOutSEN = TIMEOUT_MS_SENSOR;
		while ((startBytes[0] != 'S' || startBytes[1] != 'T')
				&& TimeOutSEN)
		{
			// Shift bytes
			startBytes[0] = startBytes[1];
			// Allow AVR time to fill transmit buffer
			Delay(1);
			// Send dummy byte to clock a response
			SendChar_SPI2(0xFF);
			// Receive byte
			startBytes[1] = ReceiveChar_SPI2();
		}

		// Receive ID byte
		sensorType[i] = ReceiveChar_SPI2();

		// Receive stop bytes
		b = 0;
		while (b < 2)
		{
			// Allow AVR time to fill transmit buffer
			Delay(1);
			// Send dummy byte to clock a response
			SendChar_SPI2(0xFF);
			// Receive byte
			stopBytes[b++] = ReceiveChar_SPI2();
		}

		// Check START pattern
		if (startBytes[0] != 'S'
				|| startBytes[1] != 'T')
			strcpy((char*)sensorData[i], "ERR");

		// Check STOP pattern
		if (stopBytes[0] != 'T'
				|| stopBytes[1] != 'S')
			strcpy((char*)sensorData[i], "ERR");

		// Check if sensor module recognised/connected
		if (sensorType[i] != 'P'
				&& sensorType[i] != 'T')
			sensorType[i] = 0;	// Sensor module not recognised/not connected!

		//Release all sensors
		SEN_Release();
	}

	return 0;
}
int ConfigureBluetooth()
{
	int proceed;
	bt_configured = 0;

	bt_frame[0] = 0;
	bt_frame[1] = 0;

	#ifdef DEBUG_USE_USART1
		SendString_USART1((uint8_t*)"\r\n-----------------------\r\nBluetooth Configuration\r\n-----------------------\r\n");
	#endif

	/* Reset BT_WAKE_HW (PC14), BT_WK_SW (PC3), BT_CMD (PC2) */
	PinReset(GPIOC, GPIO_Pin_14);
	PinReset(GPIOC, GPIO_Pin_3);
	PinReset(GPIOC, GPIO_Pin_2);
	Delay(100);

	/* Pull BT_WK_SW (PC3) high (software wake-up) */
	#ifdef DEBUG_USE_USART1
		SendString_USART1((uint8_t*)"WAKE_SW pulled high.  Waking up...");
	#endif

	ClearBuffer_USART2();
	PinSet(GPIOC, GPIO_Pin_3);
	ReceiveChars_USART2(1);	// wait for start of response
	Delay(50); 				// wait for whole response

	proceed=0;
	if (buffer_usart2[0]=='C')
		if (buffer_usart2[1]=='M')
			if (buffer_usart2[2]=='D')
				proceed=1;
	if (!proceed)
	{
		#ifdef DEBUG_USE_USART1
			SendString_USART1((uint8_t*)"Failed!\r\n");
		#endif
		return 1;
	}
	#ifdef DEBUG_USE_USART1
		SendString_USART1((uint8_t*)"Done.\r\n");
	#endif

	/* Initialise Bluetooth module */
	#ifdef DEBUG_USE_USART1
		SendString_USART1((uint8_t*)"Configuring...\r\n");
	#endif

	ClearBuffer_USART2();
	SendString_USART2((uint8_t*)"+\r");	// Echo On
	ReceiveChars_USART2(1);	// wait for start of response
	Delay(50); 				// wait for whole response
	#ifdef DEBUG_USE_USART1
		SendString_USART1(buffer_usart2);
	#endif

	ClearBuffer_USART2();
	SendString_USART2((uint8_t*)"SF,1\r");	// Factory Reset (Partial)
	ReceiveChars_USART2(1);	// wait for start of response
	Delay(50); 				// wait for whole response
	#ifdef DEBUG_USE_USART1
		SendString_USART1(buffer_usart2);
	#endif

	ClearBuffer_USART2();
	SendString_USART2((uint8_t*)"SS,00000001\r");	// Set server services = User Defined Private Profile
	ReceiveChars_USART2(1);	// wait for start of response
	Delay(50); 				// wait for whole response
	#ifdef DEBUG_USE_USART1
		SendString_USART1(buffer_usart2);
	#endif

	ClearBuffer_USART2();
	//SendString_USART2((uint8_t*)"SR,10000000\r");
	SendString_USART2((uint8_t*)"SR,30000000\r");	// Set features = Auto Advertise & Support MLDP
	ReceiveChars_USART2(1);	// wait for start of response
	Delay(50); 				// wait for whole response
	#ifdef DEBUG_USE_USART1
		SendString_USART1(buffer_usart2);
	#endif

	ClearBuffer_USART2();
	SendString_USART2((uint8_t*)"R,1\r");	// Reboot
	ReceiveChars_USART2(1);	// wait for start of response
	Delay(50); 				// wait for whole response
	#ifdef DEBUG_USE_USART1
		SendString_USART1(buffer_usart2);
	#endif

	// Receive '\0' as device goes down
	ClearBuffer_USART2();
	ReceiveChars_USART2(1);

	// Wait for device to come back up
	#ifdef DEBUG_USE_USART1
		SendString_USART1((uint8_t*)"...");
	#endif
	ClearBuffer_USART2();
	ReceiveChars_USART2(1);	// wait for start of response
	Delay(50); 				// wait for whole response

	proceed=0;
	if (buffer_usart2[0]=='C')
		if (buffer_usart2[1]=='M')
			if (buffer_usart2[2]=='D')
				proceed=1;
	if (!proceed)
	{
		#ifdef DEBUG_USE_USART1
			SendString_USART1((uint8_t*)"Failed!\r\n");
		#endif
		return 1;
	}
	#ifdef DEBUG_USE_USART1
		SendString_USART1((uint8_t*)"Done.\r\n");
	#endif

	ClearBuffer_USART2();
	SendString_USART2((uint8_t*)"I\r");	// Enter MLDP mode
	ReceiveChars_USART2(1);	// wait for start of response
	Delay(50); 				// wait for whole response
	#ifdef DEBUG_USE_USART1
		SendString_USART1(buffer_usart2);
	#endif

	#ifdef DEBUG_USE_USART1
		SendString_USART1((uint8_t*)"\r\n\r\n");
	#endif

	bt_configured = 1;
	ClearBuffer_USART2();

	return 0;
}

/******************************************************************************/

void PinSet(GPIO_TypeDef* GPIO_PORT, uint16_t GPIO_PIN)
{
	GPIO_PORT->BSRRL = GPIO_PIN;
	return;
}


void PinReset(GPIO_TypeDef* GPIO_PORT, uint16_t GPIO_PIN)
{
	GPIO_PORT->BSRRH = GPIO_PIN;
	return;
}


uint16_t PortIDR(GPIO_TypeDef* GPIO_PORT)
{
	return GPIO_PORT->IDR;
}

/******************************************************************************/

#ifdef DEBUG_USE_USART1
int SendChar_USART1(uint16_t data)
{
	/*
	 * 1) USART_WordLength_8b requires argument type uint8_t
	 * 2) uint16_t argument type used so that:
	 *    b) This function supports USART_WordLength_9b
	 *    a) This function is consistent with stm32l1xx_usart.h/c
	 * 3) Exit codes:
	 *     - 0	OK
	 *     - 1	Timed Out
	 * */

	TimeOut = TIMEOUT_MS_USART;	// Decremented by SysTick_Handler()
	while ((USART_GetFlagStatus(USART1, USART_FLAG_TXE)==RESET) && (TimeOut));

	if (TimeOut==0)
			return 1;	// Timed Out!

	USART_SendData(USART1, data);
	return 0;
}
#endif

int SendChar_USART2(uint16_t data)
{
	/*
	 * 1) USART_WordLength_8b requires argument type uint8_t
	 * 2) uint16_t argument type used so that:
	 *    b) This function supports USART_WordLength_9b
	 *    a) This function is consistent with stm32l1xx_usart.h/c
	 * 3) Exit codes:
	 *     - 0	OK
	 *     - 1	Timed Out
	 * */

	TimeOut = TIMEOUT_MS_USART;	// Decremented by SysTick_Handler()
	while ((USART_GetFlagStatus(USART2, USART_FLAG_TXE)==RESET) && (TimeOut));

	if (TimeOut==0)
			return 1;	// Timed Out!

	USART_SendData(USART2, data);
	return 0;
}

int SendChar_SPI2(uint16_t data)
{
	/*
	 * 1) SPI_DataSize_8b requires argument type uint8_t
	 * 2) uint16_t argument type used so that:
	 *    b) This function supports SPI_DataSize_16b
	 *    a) This function is consistent with stm32l1xx_spi.h/c
	 * 3) Exit codes:
	 *     - 0	OK
	 *     - 1	Timed Out
	 * */

	TimeOut = TIMEOUT_MS_SPI;	// Decremented by SysTick_Handler()
	while ((SPI_I2S_GetFlagStatus(SPI2, SPI_I2S_FLAG_TXE)==RESET) && (TimeOut));

	if (TimeOut==0)
		return 1;	// Timed Out!

	SPI_I2S_SendData(SPI2, data);
	return 0;
}

int SendChar_SPI3(uint16_t data)
{
	/*
	 * 1) SPI_DataSize_8b requires argument type uint8_t
	 * 2) uint16_t argument type used so that:
	 *    b) This function supports SPI_DataSize_16b
	 *    a) This function is consistent with stm32l1xx_spi.h/c
	 * 3) Exit codes:
	 *     - 0	OK
	 *     - 1	Timed Out
	 * */

	TimeOut = TIMEOUT_MS_SPI;	// Decremented by SysTick_Handler()
	while ((SPI_I2S_GetFlagStatus(SPI3, SPI_I2S_FLAG_TXE)==RESET) && (TimeOut));

	if (TimeOut==0)
		return 1;	// Timed Out!

	SPI_I2S_SendData(SPI3, data);
	return 0;
}

int SendChar_I2C2(uint8_t reg_addr, uint8_t reg_data)
{
	I2C_GenerateSTART(I2C2, ENABLE);
	while(!I2C_CheckEvent(I2C2, I2C_EVENT_MASTER_MODE_SELECT));

	I2C_Send7bitAddress(I2C2, RTC_DEVICE_ADDR, I2C_Direction_Transmitter);
	while(!I2C_CheckEvent(I2C2, I2C_EVENT_MASTER_TRANSMITTER_MODE_SELECTED));

	I2C_SendData(I2C2, reg_addr);
	while(!I2C_CheckEvent(I2C2, I2C_EVENT_MASTER_BYTE_TRANSMITTED));

	I2C_SendData(I2C2, reg_data);
	while(!I2C_CheckEvent(I2C2, I2C_EVENT_MASTER_BYTE_TRANSMITTED));

	Delay(1);
	I2C_GenerateSTOP(I2C2, ENABLE);
	Delay(1);

	return 0;
}

/******************************************************************************/

#ifdef DEBUG_USE_USART1
uint16_t ReceiveChar_USART1()
{
	/*
	 * 1) USART_WordLength_8b requires return type uint8_t
	 * 2) uint16_t return type used so that:
	 *    b) This function supports USART_WordLength_9b
	 *    a) This function is consistent with stm32l1xx_usart.h/c
	 *    c) The upper byte may be used to convey additional information:
	 *       (When using USART_WordLength_8b)
	 *       - Exit code 1000 = Timed Out
	 * */

	TimeOut = TIMEOUT_MS_USART;	// Decremented by SysTick_Handler()
	while ((USART_GetFlagStatus(USART1, USART_FLAG_RXNE)==RESET) && (TimeOut));

	if (TimeOut==0)
		return 1000;	// Timed Out!

	return USART_ReceiveData(USART1);
}
#endif

uint16_t ReceiveChar_USART2()
{
	/*
	 * 1) USART_WordLength_8b requires return type uint8_t
	 * 2) uint16_t return type used so that:
	 *    b) This function supports USART_WordLength_9b
	 *    a) This function is consistent with stm32l1xx_usart.h/c
	 *    c) The upper byte may be used to convey additional information:
	 *       (When using USART_WordLength_8b)
	 *       - Exit code 1000 = Timed Out
	 * */

	TimeOut = TIMEOUT_MS_USART;	// Decremented by SysTick_Handler()
	while ((USART_GetFlagStatus(USART2, USART_FLAG_RXNE)==RESET) && (TimeOut));

	if (TimeOut==0)
		return 1000;	// Timed Out!

	return USART_ReceiveData(USART2);
}

uint16_t ReceiveChar_SPI2()
{
	/*
	 * 1) SPI_DataSize_8b requires return type uint8_t
	 * 2) uint16_t return type used so that:
	 *    b) This function supports SPI_DataSize_16b
	 *    a) This function is consistent with stm32l1xx_spi.h/c
	 *    c) The upper byte may be used to convey additional information:
	 *       (When using SPI_DataSize_8b)
	 *       - Exit code 1000 = Timed Out
	 * */

	TimeOut = TIMEOUT_MS_SPI;	// Decremented by SysTick_Handler()
	while ((SPI_I2S_GetFlagStatus(SPI2, SPI_I2S_FLAG_RXNE)==RESET) && (TimeOut!=0));

	if (TimeOut==0)
		return 1000;	// Timed Out!

	return SPI_I2S_ReceiveData(SPI2);
}

uint16_t ReceiveChar_SPI3()
{
	/*
	 * 1) SPI_DataSize_8b requires return type uint8_t
	 * 2) uint16_t return type used so that:
	 *    b) This function supports SPI_DataSize_16b
	 *    a) This function is consistent with stm32l1xx_spi.h/c
	 *    c) The upper byte may be used to convey additional information:
	 *       (When using SPI_DataSize_8b)
	 *       - Exit code 1000 = Timed Out
	 * */

	TimeOut = TIMEOUT_MS_SPI;	// Decremented by SysTick_Handler()
	while ((SPI_I2S_GetFlagStatus(SPI3, SPI_I2S_FLAG_RXNE)==RESET) && (TimeOut));

	if (TimeOut==0)
		return 1000;	// Timed Out!

	return SPI_I2S_ReceiveData(SPI3);
}

uint8_t ReceiveChar_I2C2(uint8_t reg_addr)
{
	uint8_t data;

	I2C_GenerateSTART(I2C2, ENABLE);
	while(!I2C_CheckEvent(I2C2, I2C_EVENT_MASTER_MODE_SELECT));

	I2C_Send7bitAddress(I2C2, RTC_DEVICE_ADDR, I2C_Direction_Transmitter);
	while(!I2C_CheckEvent(I2C2, I2C_EVENT_MASTER_TRANSMITTER_MODE_SELECTED));

	I2C_SendData(I2C2, reg_addr);
	while(!I2C_CheckEvent(I2C2, I2C_EVENT_MASTER_BYTE_TRANSMITTED));

	I2C_GenerateSTART(I2C2, ENABLE);
	while(!I2C_CheckEvent(I2C2, I2C_EVENT_MASTER_MODE_SELECT));

	I2C_Send7bitAddress(I2C2, RTC_DEVICE_ADDR, I2C_Direction_Receiver);
	while(!I2C_CheckEvent(I2C2, I2C_EVENT_MASTER_RECEIVER_MODE_SELECTED));

	while(!I2C_CheckEvent(I2C2, I2C_EVENT_MASTER_BYTE_RECEIVED));
	data = I2C_ReceiveData(I2C2);

	Delay(1);
	I2C_GenerateSTOP(I2C2, ENABLE);
	Delay(1);

	return data;
}

/******************************************************************************/

#ifdef DEBUG_USE_USART1
void SendString_USART1(uint8_t data[])
{
	uint16_t i=0;
	while (data[i]!='\0')	// loop until null termination character '\0'
	{
		while (!USART_GetFlagStatus(USART1, USART_FLAG_TXE));	// while !(transmit data register empty flag)
		USART_SendData(USART1, data[i++]);
	}
	return;
}
#endif

void SendString_USART2(uint8_t data[])
{
	uint16_t i=0;
	while (data[i]!='\0')	// loop until null termination character '\0'
	{
		while (!USART_GetFlagStatus(USART2, USART_FLAG_TXE));	// while !(transmit data register empty flag)
		USART_SendData(USART2, data[i++]);
	}
	return;
}

#ifdef DEBUG_USE_USART1
void SendChars_USART1(uint8_t data[], int numChars)
{
	uint16_t i=0;
	while (data[i]!='\0')	// loop until null termination character '\0'
	{
		while (!USART_GetFlagStatus(USART1, USART_FLAG_TXE));	// while !(transmit data register empty flag)
		USART_SendData(USART1, data[i++]);

		if (i>=numChars)
			break;
	}
	return;
}
#endif

void SendChars_USART2(uint8_t data[], int numChars)
{
	uint16_t i=0;
	while (data[i]!='\0')	// loop until null termination character '\0'
	{
		while (!USART_GetFlagStatus(USART2, USART_FLAG_TXE));	// while !(transmit data register empty flag)
		USART_SendData(USART2, data[i++]);

		if (i>=numChars)
			break;
	}
	return;
}

#ifdef DEBUG_USE_USART1
int ReceiveString_USART1()
{
	while(rx_index_usart1<MAX_CHARS)
	{
		if(buffer_usart1[rx_index_usart1]=='\r')
			break;
	}

	if(rx_index_usart1>=MAX_CHARS)
		return 1;

	return 0;
}
#endif

int ReceiveString_USART2()
{
	while(rx_index_usart2<MAX_CHARS)
	{
		if(buffer_usart2[rx_index_usart2]=='\r')
			break;
	}

	if(rx_index_usart2>=MAX_CHARS)
		return 1;

	return 0;
}

#ifdef DEBUG_USE_USART1
int ReceiveChars_USART1(int numChars)
{
	while(rx_index_usart1<MAX_CHARS)
	{
		if(rx_index_usart1>numChars-1)
			break;
		//if(buffer_usart1[rx_index_usart1]=='\r')
			//break;
	}

	if(rx_index_usart1>=MAX_CHARS)
		return 1;

	return 0;
}
#endif

int ReceiveChars_USART2(int numChars)
{
	while(rx_index_usart2<MAX_CHARS)
	{
		if(rx_index_usart2>numChars-1)
			break;
		//if(buffer_usart2[rx_index_usart2]=='\r')
			//break;
	}

	if(rx_index_usart2>=MAX_CHARS)
		return 1;

	return 0;
}

/******************************************************************************/

/* User implemented function for fatfs/diskio */
#if !_FS_NORTC && !_FS_READONLY
DWORD get_fattime (void)
{
	return	  ((DWORD)(I2C_RTC_GetYear() + 20) << 25)
			| ((DWORD)I2C_RTC_GetMonth() << 21)
			| ((DWORD)I2C_RTC_GetDate() << 16)
			| ((DWORD)I2C_RTC_GetHours() << 11)
			| ((DWORD)I2C_RTC_GetMinutes() << 5)
			| ((DWORD)I2C_RTC_GetSeconds() >> 1);
}
#endif

void I2C_RTC_Buffer()
{
	uint8_t day = I2C_RTC_GetDay();

	if ((day>0) && (day<=7))
		buffer_day = dow[day-1];

	uint8_t date = I2C_RTC_GetDate();
	uint8_t month = I2C_RTC_GetMonth();
	uint8_t year = I2C_RTC_GetYear();

	buffer_date[0] = '0' + (date / 10);
	buffer_date[1] = '0' + (date % 10);
	buffer_date[2] = date_separator;
	buffer_date[3] = '0' + (month / 10);
	buffer_date[4] = '0' + (month % 10);
	buffer_date[5] = date_separator;
	buffer_date[6] = '2';
	buffer_date[7] = '0';
	buffer_date[8] = '0' + (year / 10);
	buffer_date[9] = '0' + (year % 10);

	uint8_t hours = I2C_RTC_GetHours();
	uint8_t minutes = I2C_RTC_GetMinutes();
	uint8_t seconds = I2C_RTC_GetSeconds();

	buffer_time[0] = '0' + (hours / 10);
	buffer_time[1] = '0' + (hours % 10);
	buffer_time[2] = time_separator;
	buffer_time[3] = '0' + (minutes / 10);
	buffer_time[4] = '0' + (minutes % 10);
	buffer_time[5] = time_separator;
	buffer_time[6] = '0' + (seconds / 10);
	buffer_time[7] = '0' + (seconds % 10);

	return;
}

uint8_t I2C_RTC_GetSeconds()
{
	uint8_t rtc_data = ReceiveChar_I2C2(RTC_ADDR_SECONDS);

	uint8_t tens = (rtc_data & 0x70) >> 4;
	uint8_t units = (rtc_data & 0x0F);

	uint8_t total = (tens*10) + units;

	return total;
}
uint8_t I2C_RTC_GetMinutes()
{
	uint8_t rtc_data = ReceiveChar_I2C2(RTC_ADDR_MINUTES);

	uint8_t tens = (rtc_data & 0x70) >> 4;
	uint8_t units = (rtc_data & 0x0F);

	uint8_t total = (tens*10) + units;

	return total;
}
uint8_t I2C_RTC_GetHours()
{
	uint8_t rtc_data = ReceiveChar_I2C2(RTC_ADDR_HOURS);

	uint8_t tens = (rtc_data & 0x30) >> 4;
	uint8_t units = (rtc_data & 0x0F);

	uint8_t total = (tens*10) + units;

	return total;
}
uint8_t I2C_RTC_GetDay()
{
	uint8_t rtc_data = ReceiveChar_I2C2(RTC_ADDR_DAY);

	uint8_t day = (rtc_data & 0x07);

	return day;
}
uint8_t I2C_RTC_GetDate()
{
	uint8_t rtc_data = ReceiveChar_I2C2(RTC_ADDR_DATE);

	uint8_t tens = (rtc_data & 0x30) >> 4;
	uint8_t units = (rtc_data & 0x0F);

	uint8_t total = (tens*10) + units;

	return total;
}
uint8_t I2C_RTC_GetMonth()
{
	uint8_t rtc_data = ReceiveChar_I2C2(RTC_ADDR_MONTH);

	uint8_t tens = (rtc_data & 0x10) >> 4;
	uint8_t units = (rtc_data & 0x0F);

	uint8_t total = (tens*10) + units;

	return total;
}
uint8_t I2C_RTC_GetYear()
{
	uint8_t rtc_data = ReceiveChar_I2C2(RTC_ADDR_YEARS);

	uint8_t tens = (rtc_data & 0xF0) >> 4;
	uint8_t units = (rtc_data & 0x0F);

	uint8_t total = (tens*10) + units;

	return total;
}


void I2C_RTC_SetSeconds(uint8_t value)
{
	if ((value<0) && (value>59))
		return;

	uint8_t tens = value / 10;
	uint8_t units = value % 10;
	uint8_t total = (tens << 4) + units;

	SendChar_I2C2(RTC_ADDR_SECONDS, total);
	return;
}
void I2C_RTC_SetMinutes(uint8_t value)
{
	if ((value<0) && (value>59))
		return;

	uint8_t tens = value / 10;
	uint8_t units = value % 10;
	uint8_t total = (tens << 4) + units;

	SendChar_I2C2(RTC_ADDR_MINUTES, total);
	return;
}
void I2C_RTC_SetHours(uint8_t value)
{
	if ((value<0) && (value>23))
		return;

	uint8_t tens = value / 10;
	uint8_t units = value % 10;
	uint8_t total = (tens << 4) + units;

	SendChar_I2C2(RTC_ADDR_HOURS, total);
	return;
}
void I2C_RTC_SetDay(uint8_t value)
{
	if ((value<1) && (value>7))
		return;

	SendChar_I2C2(RTC_ADDR_DAY, value);
	return;
}
void I2C_RTC_SetDate(uint8_t value)
{
	if ((value<0) && (value>31))
		return;

	uint8_t tens = value / 10;
	uint8_t units = value % 10;
	uint8_t total = (tens << 4) + units;

	SendChar_I2C2(RTC_ADDR_DATE, total);
	return;
}
void I2C_RTC_SetMonth(uint8_t value)
{
	if ((value<0) && (value>12))
		return;

	uint8_t tens = value / 10;
	uint8_t units = value % 10;
	uint8_t total = (tens << 4) + units;

	SendChar_I2C2(RTC_ADDR_MONTH, total);
	return;
}
void I2C_RTC_SetYear(uint8_t value)
{
	if ((value<0) && (value>99))
		return;

	uint8_t tens = value / 10;
	uint8_t units = value % 10;
	uint8_t total = (tens << 4) + units;

	SendChar_I2C2(RTC_ADDR_YEARS, total);
	return;
}


void I2C_RTC_Start()
{
	/* Reset STOP bit (MSB) in SECONDS register */
	uint8_t rtc_data = ReceiveChar_I2C2(RTC_ADDR_SECONDS);
	rtc_data &= ~(0x80);
	SendChar_I2C2(RTC_ADDR_SECONDS, rtc_data);	// reset STOP bit

	return;
}
void I2C_RTC_Stop()
{
	/* Set STOP bit (MSB) in SECONDS register */
	uint8_t rtc_data = ReceiveChar_I2C2(RTC_ADDR_SECONDS);
	rtc_data |= 0x80;
	SendChar_I2C2(RTC_ADDR_SECONDS, rtc_data);	// set STOP bit

	return;
}


void I2C_RTC_ResetDayDateTime()
{
	SendChar_I2C2(RTC_ADDR_SECONDS, 0x00);
	SendChar_I2C2(RTC_ADDR_MINUTES, 0x00);
	SendChar_I2C2(RTC_ADDR_HOURS, 0x00);
	SendChar_I2C2(RTC_ADDR_DAY, 0x00);
	SendChar_I2C2(RTC_ADDR_DATE, 0x00);
	SendChar_I2C2(RTC_ADDR_MONTH, 0x00);
	SendChar_I2C2(RTC_ADDR_YEARS, 0x00);
}

/******************************************************************************/

void RTC_Set(char* dateString)
{
	// Format= {DoW <10 chars}.dd-MM-yy.HH:mm:ss

	int i, tens, units;
	uint8_t tempDow[10];
	for (i = 0; i < 10; i++)
		tempDow[i] = 0;

	// Get DoW
	i = 0;
	while (dateString[i] != '.'				// DoW delimiter
			&& i < 10						// protect tempDow overflow
			&& dateString[i] != '\0')		// check end of dateString
	{
		tempDow[i] = dateString[i];
		i++;
	}

	if ((strlen(dateString) - i - 1) < 17
			|| (dateString[i] != '.' && dateString[i] != '\0'))
		return;	// invalid dateString
	i++;

	// Set Date
	tens = dateString[i++] - '0';
	units = dateString[i++] - '0';
	I2C_RTC_SetDate((10*tens)+units);
	i++;

	tens = dateString[i++] - '0';
	units = dateString[i++] - '0';
	I2C_RTC_SetMonth((10*tens)+units);
	i++;

	tens = dateString[i++] - '0';
	units = dateString[i++] - '0';
	I2C_RTC_SetYear((10*tens)+units);
	i++;

	// Set Time
	tens = dateString[i++] - '0';
	units = dateString[i++] - '0';
	I2C_RTC_SetHours((10*tens)+units);
	i++;

	tens = dateString[i++] - '0';
	units = dateString[i++] - '0';
	I2C_RTC_SetMinutes((10*tens)+units);
	i++;

	tens = dateString[i++] - '0';
	units = dateString[i++] - '0';
	I2C_RTC_SetSeconds((10*tens)+units);

	// Set DoW
	for (i = 0; i < 7; i++)
	{
		if (strcmp((char*)tempDow, (char*)dow[i]) == 0)
		{
			I2C_RTC_SetDay(i);
			break;
		}
	}
}

/******************************************************************************/

void CheckCmd(char* cmdStr)
{
	switch (cmdStr[0])
	{
		case 'T':		// "T?" = beginning of time set (where ?=any char delimiter)
			// Format= {DoW <10 chars}.dd-MM-yy.HH:mm:ss
			RTC_Set(cmdStr+2);
			break;
		case 'Y':
			if (!log_run)
				Logging_Start();
			break;
		case 'N':
			if (log_run)
				Logging_Stop();
			break;
		case 'L':
			bt_live = 1;
			break;
		case 'K':
			bt_live = 0;
			break;
		default:
			break;
	}
}

void BT_Check()
{
	bt_received = 0;	//reset flag

	//DEBUG check code not dead
//	SendString_USART1((uint8_t*)"BT_Check(): ");
//	SendString_USART1(buffer_usart2);
//	SendString_USART1((uint8_t*)"\n\r");

	CheckCmd((char*)buffer_usart2);
}

/******************************************************************************/

void SEN_Select(int sensor)
{
	SEN_Release();

	switch(sensor)
	{
		case 0:
			PinReset(GPIOA, GPIO_Pin_11);	// SEN_1_SS (PA11)
			break;
		case 1:
			PinReset(GPIOA, GPIO_Pin_9);	// SEN_2_SS (PA9)
			break;
		case 2:
			PinReset(GPIOA, GPIO_Pin_8);	// SEN_3_SS (PA8)
			break;
		case 3:
			PinReset(GPIOA, GPIO_Pin_10);	// SEN_4_SS (PA10)
			break;
		default:
			break;
	}
}

void SEN_Release()
{
	PinSet(GPIOA, GPIO_Pin_11);		// SEN_1_SS (PA11)
	PinSet(GPIOA, GPIO_Pin_9);		// SEN_2_SS (PA9)
	PinSet(GPIOA, GPIO_Pin_8);		// SEN_3_SS (PA8)
	PinSet(GPIOA, GPIO_Pin_10);		// SEN_4_SS (PA10)
}

void SEN_GetData(void)
{
	uint8_t startBytes[2];
	uint8_t stopBytes[2];
	uint8_t dataBytes = 0;

	int i, b;
	for (i = 0; i < 4; i++)
	{
		startBytes[0] = 0;
		startBytes[1] = 0;
		stopBytes[0] = 0;
		stopBytes[1] = 0;

		switch (sensorType[i])
		{
			case 'P':
				#ifdef DEBUG_USE_USART1
					SendString_USART1((uint8_t*)"Requesting P Data...\n\r");
				#endif

				dataBytes = SENSOR_P_DATA_BYTES;
				break;

			case 'T':
				#ifdef DEBUG_USE_USART1
					SendString_USART1((uint8_t*)"Requesting T Data...\n\r");
				#endif

				dataBytes = SENSOR_T_DATA_BYTES;
				break;

			case 0:
				// No sensor module connected!
				strcpy((char*)sensorData[i], "NC");
				continue;

			default:
				// Error!
				strcpy((char*)sensorData[i], "ERR");
				break;
		}

		// Select Sensor i
		SEN_Select(i);

		// Request data
		SendChar_SPI2('R');

		// Wait for start bytes
		TimeOutSEN = TIMEOUT_MS_SENSOR;
		while ((startBytes[0] != 'S' || startBytes[1] != 'T')
				&& TimeOutSEN)
		{
			// Shift bytes
			startBytes[0] = startBytes[1];
			// Allow AVR time to fill transmit buffer
			Delay(1);
			// Send dummy byte to clock a response
			SendChar_SPI2(0xFF);
			// Receive byte
			startBytes[1] = ReceiveChar_SPI2();
		}
		// Receive data bytes
		b = 0;
		while(b < dataBytes)
		{
			// Allow AVR time to fill transmit buffer
			Delay(1);
			// Send dummy byte to clock a response
			SendChar_SPI2(0xFF);
			// Receive byte
			sensorData[i][b++] = ReceiveChar_SPI2();
		}
		// Null termination byte
		sensorData[i][b] = 0;
		// Receive stop bytes
		b = 0;
		while (b < 2)
		{
			// Allow AVR time to fill transmit buffer
			Delay(1);
			// Send dummy byte to clock a response
			SendChar_SPI2(0xFF);
			// Receive byte
			stopBytes[b++] = ReceiveChar_SPI2();
		}

		// Check START pattern
		if (startBytes[0] != 'S'
				|| startBytes[1] != 'T')
			strcpy((char*)sensorData[i], "ERR");

		// Check STOP pattern
		if (stopBytes[0] != 'T'
				|| stopBytes[1] != 'S')
			strcpy((char*)sensorData[i], "ERR");

		//Release all sensors
		SEN_Release();
	}
}

void SEN_FormatData()
{
	int i, j;

	// Clear sensorLogData
	sensorLogData[0] = 0;
	// Format sensorData to be logged
	for (i = 0; i < 4; i++)
	{
		switch (sensorType[i])
		{
			case 0:
				// Sensor module Not Connected (NC)
				sprintf(sensorLogData, "%s%s,", sensorLogData, "NC");
				break;

			case 'P':
				// Write sensor type
				sprintf(sensorLogData, "%s%c,", sensorLogData, sensorType[i]);
				// Check for data error
				if (strcmp((char*)sensorData[i], "ERR") == 0)
				{
					sprintf(sensorLogData, "%s%s,", sensorLogData, "ERR");
					break;
				}
				// Write data
				for (j = 0; j < 7; j++)
				{
					joinData = sensorData[i][(2*j)+1] << 8;
					joinData |= sensorData[i][(2*j)];
					sprintf(sensorLogData, "%s%u,", sensorLogData, joinData);
				}
				joinData = sensorData[i][(2*j)];
				sprintf(sensorLogData, "%s%u,", sensorLogData, joinData);
				break;

			case 'T':
				// Write sensor type
				sprintf(sensorLogData, "%s%c,", sensorLogData, sensorType[i]);
				// Check for data error
				if (strcmp((char*)sensorData[i], "ERR") == 0)
				{
					sprintf(sensorLogData, "%s%s,", sensorLogData, "ERR");
					break;
				}
				// Write data
				for (j = 0; j < 4; j++)
				{
					joinData = sensorData[i][(4*j)+3] << 24;
					joinData |= sensorData[i][(4*j)+2] << 16;
					joinData |= sensorData[i][(4*j)+1] << 8;
					joinData |= sensorData[i][(4*j)];
					sprintf(sensorLogData, "%s%u,", sensorLogData, joinData);
				}
				break;

			default:
				break;
		}
	}
	// Remove trailing ','
	sensorLogData[strlen(sensorLogData) - 1] = 0;
	// Add '\n'
	sprintf(sensorLogData, "%s%c", sensorLogData, '\n');
}

/******************************************************************************/

void SD_CS_High()
{
	PinSet(GPIOB, GPIO_Pin_4);
}

void SD_CS_Low()
{
	PinReset(GPIOB, GPIO_Pin_4);
}

void FRESULTtoString(FRESULT fr)
{
	const char *str =
		"FR_OK\0" "FR_DISK_ERR\0" "FR_INT_ERR\0" "FR_NOT_READY\0" "FR_NO_FILE\0"
		"FR_NO_PATH\0" "FR_INVALID_NAME\0" "FR_DENIED\0" "FR_EXIST\0" "FR_INVALID_OBJECT\0"
		"FR_WRITE_PROTECTED\0" "FR_INVALID_DRIVE\0" "FR_NOT_ENABLED\0" "FR_NO_FILESYSTEM\0" "FR_MKFS_ABORTED\0"
		"FR_TIMEOUT\0" "FR_LOCKED\0" "FR_NOT_ENOUGH_CORE\0" "FR_TOO_MANY_OPEN_FILES\0" "FR_INVALID_PARAMETER\0";
	FRESULT i;

	for (i = 0; i != fr && *str; i++) {
		while (*str++) ;
	}

	strcpy(frbuff, str);
}

/******************************************************************************/

int SD_Mount()
{
	/*
	 * 0 = OK
	 * 1 = Failed to mount SD card!
	 * 2 = Insert SD card!
	 * 3 = Remove SD card lock!
	 * 4 = Invalid DSTATUS!
	 *
	 * */

	DSTATUS dstatus;
	FRESULT fresult;
	char dstatusStr[4];	// dstatus = uint8_t with maximum value 256

	/* Allow SD card to wake-up */
	Delay(500);

	/* Check disk status */
	dstatus = disk_status(0);
	#ifdef DEBUG_USE_USART1
		SendString_USART1((uint8_t*)"DSTATUS:\n\r");
		if (dstatus)
		{
			if (dstatus & STA_NOINIT)
				SendString_USART1((uint8_t*)"  STA_NOINIT\n\r");
			if (dstatus & STA_NODISK)
				SendString_USART1((uint8_t*)"  STA_NODISK\n\r");
			if (dstatus & STA_PROTECT)
				SendString_USART1((uint8_t*)"  STA_PROTECT\n\r");
		}
		else
			SendString_USART1((uint8_t*)"  OK\n\r");
	#endif

	if(!dstatus)
	{
		#ifdef DEBUG_USE_USART1
			SendString_USART1((uint8_t*)"SD card already mounted.\n\r");
		#endif
		return 0;
	}
	else
	{
		if (dstatus & STA_NODISK)
		{
			#ifdef DEBUG_USE_USART1
				SendString_USART1((uint8_t*)"Insert SD card!\n\r");
			#endif
			return 2;
		}
		else if (dstatus & STA_PROTECT)
		{
			#ifdef DEBUG_USE_USART1
				SendString_USART1((uint8_t*)"Remove SD card lock!\n\r");
			#endif
			return 3;
		}
		else if (dstatus & STA_NOINIT)
		{
			#ifdef DEBUG_USE_USART1
				SendString_USART1((uint8_t*)"Mounting SD card...");
			#endif

			fresult = f_mount(&fs, "", 1);

			#ifdef DEBUG_USE_USART1
				FRESULTtoString(fresult);
				SendString_USART1((uint8_t*)frbuff);
				SendString_USART1((uint8_t*)"\n\r");
			#endif

			if (fresult == FR_OK)
			{
				#ifdef DEBUG_USE_USART1
					SendString_USART1((uint8_t*)"SD card mounted.\n\r");
				#endif
				return 0;
			}
			else
			{
				#ifdef DEBUG_USE_USART1
					SendString_USART1((uint8_t*)"Failed to mount SD card!\n\r");
				#endif
				return 1;
			}
		}
		else
		{
			#ifdef DEBUG_USE_USART1
				snprintf(dstatusStr, 4, "%u", dstatus);

				SendString_USART1((uint8_t*)"Invalid DSTATUS!\n\r");
				SendString_USART1((uint8_t*)dstatusStr);
				SendString_USART1((uint8_t*)"\n\r");
			#endif
			return 4;
		}
	}
}

void SD_Unmount()
{
	sd_isr_unmount = 0;	// reset flag in case SD_Unmount() was triggered by ISR

    f_mount(NULL, "0:", 0);
    sd_mounted = 0;

	return;
}

int SD_OpenConfig()
{
	FRESULT fresult;

	/* Check SD card mounted */
	if (!sd_mounted)
		return 1;

	/* Set filename */
	strcpy(configFilename, "CONFIG.TXT");

	/* Open file */
	#ifdef DEBUG_USE_USART1
		SendString_USART1((uint8_t*)"Opening ");
		SendString_USART1((uint8_t*)configFilename);
		SendString_USART1((uint8_t*)"\n\r");
	#endif

	fresult = f_open(&fpConfig, (char*)configFilename, FA_OPEN_EXISTING | FA_READ);

	#ifdef DEBUG_USE_USART1
		FRESULTtoString(fresult);
		SendString_USART1((uint8_t*)frbuff);
		SendString_USART1((uint8_t*)"\n\r");
	#endif

	if (fresult == FR_OK)
	{
		#ifdef DEBUG_USE_USART1
			SendString_USART1((uint8_t*)"File open.\n\r");
		#endif
		return 0;
	}
	else
	{
		#ifdef DEBUG_USE_USART1
			SendString_USART1((uint8_t*)"Failed to open file!\n\r");
		#endif
		return 1;
	}
}

int SD_CloseConfig()
{
	FRESULT fresult;

	/* Check SD card mounted */
	if (!sd_mounted)
		return 1;

	/* Close file */
	#ifdef DEBUG_USE_USART1
		SendString_USART1((uint8_t*)"Closing ");
		SendString_USART1((uint8_t*)configFilename);
		SendString_USART1((uint8_t*)"\n\r");
	#endif

	fresult = f_close(&fpConfig);

	#ifdef DEBUG_USE_USART1
		FRESULTtoString(fresult);
		SendString_USART1((uint8_t*)frbuff);
		SendString_USART1((uint8_t*)"\n\r");
	#endif

	if (fresult == FR_OK)
	{
		#ifdef DEBUG_USE_USART1
			SendString_USART1((uint8_t*)"File closed.\n\r");
		#endif
		return 0;
	}
	else
	{
		#ifdef DEBUG_USE_USART1
			SendString_USART1((uint8_t*)"Failed to close file!\n\r");
		#endif
		return 1;
	}
}

int SD_CheckConfig()
{
	int res = 0;

	sd_isr_mount_config = 0; // reset flag in case SD_CheckConfig() was triggered by ISR (ConfigureSD())

	res = SD_OpenConfig();	// returns 0 if OK

	// Check SD card configuration file
	if (!res)
	{
		res = 0;	// return code OK
		while (!f_eof(&fpConfig) && !f_error(&fpConfig))
		{
			if(!f_gets(rbuff, MAX_CHARS+1, &fpConfig))
				break;
			CheckCmd(rbuff);
		}
	}
	else
		res = 1;	// return code NO CONFIG

	SD_CloseConfig();

	return res;
}

int SD_OpenLog()
{
	FRESULT fresult;

	sd_isr_mount_open = 0; // reset flag in case SD_OpenLog() was triggered by ISR (ConfigureSD())

	/* Check SD card mounted */
	if (!sd_mounted)
		return 1;

	/* Open file */
	#ifdef DEBUG_USE_USART1
		SendString_USART1((uint8_t*)"Opening ");
		SendString_USART1((uint8_t*)logFilename);
		SendString_USART1((uint8_t*)"\n\r");
	#endif

	fresult = f_open(&fp, (char*)logFilename, FA_OPEN_ALWAYS | FA_READ | FA_WRITE);

	#ifdef DEBUG_USE_USART1
		FRESULTtoString(fresult);
		SendString_USART1((uint8_t*)frbuff);
		SendString_USART1((uint8_t*)"\n\r");
	#endif

	if (fresult == FR_OK)
	{
		#ifdef DEBUG_USE_USART1
			SendString_USART1((uint8_t*)"File open.\n\r");
		#endif
		return 0;
	}
	else
	{
		#ifdef DEBUG_USE_USART1
			SendString_USART1((uint8_t*)"Failed to open file!\n\r");
		#endif
		return 1;
	}
}

int SD_CloseLog()
{
	FRESULT fresult;

	/* Check SD card mounted */
	if (!sd_mounted)
		return 1;

	/* Close file */
	#ifdef DEBUG_USE_USART1
		SendString_USART1((uint8_t*)"Closing ");
		SendString_USART1((uint8_t*)logFilename);
		SendString_USART1((uint8_t*)"\n\r");
	#endif

	fresult = f_close(&fp);

	#ifdef DEBUG_USE_USART1
		FRESULTtoString(fresult);
		SendString_USART1((uint8_t*)frbuff);
		SendString_USART1((uint8_t*)"\n\r");
	#endif

	if (fresult == FR_OK)
	{
		#ifdef DEBUG_USE_USART1
			SendString_USART1((uint8_t*)"File closed.\n\r");
		#endif
		return 0;
	}
	else
	{
		#ifdef DEBUG_USE_USART1
			SendString_USART1((uint8_t*)"Failed to close file!\n\r");
		#endif
		return 1;
	}
}

int SD_LogReadLine()
{
	/* Check SD card mounted */
	if (!sd_mounted)
		return 1;

	f_gets((char*)rbuff, MAX_CHARS+1, &fp);

	return 0;
}

int SD_LogWriteLine()
{
	FRESULT fresult;
	int res = 0;

	/* Check SD card mounted */
	if (!sd_mounted)
		return -2;

	/* Move to end of the file to append data */
	fresult = f_lseek(&fp, f_size(&fp));

	if (fresult == FR_OK)
	{
		/* Write wbuff string to file */
		res = f_puts(wbuff, &fp);
		if (res == -1)
		{
			#ifdef DEBUG_USE_USART1
				SendString_USART1((uint8_t*)"Failed to append - reached EOF!\n\r");
			#endif
		}
		else
		{
			#ifdef DEBUG_USE_USART1
				char resStr[6];	// res = int with maximum value 65535
				snprintf(resStr, 6, "%u", res);
				SendString_USART1((uint8_t*)"Written ");
				SendString_USART1((uint8_t*)resStr);
				SendString_USART1((uint8_t*)"chars to file.\n\r");
			#endif
		}

		/* Flush cache */
		fresult = f_sync(&fp);
		if (fresult == FR_OK)
		{
			#ifdef DEBUG_USE_USART1
				SendString_USART1((uint8_t*)"Flushed cache.\n\r");
			#endif
			return res;
		}
		else
		{
			#ifdef DEBUG_USE_USART1
				SendString_USART1((uint8_t*)"Failed to flush cache!\n\r");

				FRESULTtoString(fresult);
				SendString_USART1((uint8_t*)frbuff);
				SendString_USART1((uint8_t*)"\n\r");
			#endif
			return -2;
		}
	}
	else
	{
		#ifdef DEBUG_USE_USART1
			SendString_USART1((uint8_t*)"Failed to find EOF for append!\n\r");
		#endif
		return -3;
	}
}

/******************************************************************************/

void Logging_Start(void)
{
	log_isr_start = 0;	// reset flag in case Logging_Start() was triggered by ISR

	/* Generate logFilename */
	int i;
	char tempFilename[13];

	uint8_t date = I2C_RTC_GetDate();
	uint8_t month = I2C_RTC_GetMonth();
	uint8_t year = I2C_RTC_GetYear();
	logFilename[0] = '0' + (year / 10);
	logFilename[1] = '0' + (year % 10);
	logFilename[2] = '0' + (month / 10);
	logFilename[3] = '0' + (month % 10);
	logFilename[4] = '0' + (date / 10);
	logFilename[5] = '0' + (date % 10);
	logFilename[6] = '_';
	logFilename[7] = 0;
	logFilename[13] = 0;

	// Try appending 0...9 until file doesn't exist
	i = 0;
	sprintf(tempFilename, "%s%u%s", logFilename, i++, ".LOG");
	while (f_stat(tempFilename, &fno) != FR_NO_FILE && i < 10)
	{
		sprintf(tempFilename, "%s%u%s", logFilename, i, ".LOG");
		i++;
	}
	// if all 10 filenames taken for today's date...
	if (i == 10)
		i = 0;	// ...reset
	else
		i--;	// remove increment which broke out of loop

	// Append number + extension
	sprintf(logFilename, "%s%u%s", logFilename, i, ".LOG");

	SD_OpenLog();

	log_run = 1;
}

void Logging_Stop(void)
{
	log_isr_stop = 0;	// reset flag in case Logging_Stop() was triggered by ISR

	SD_CloseLog();

	/* Erase logFilename */
	strcpy(logFilename, "");

	log_run = 0;
}

/******************************************************************************/

#ifdef DEBUG_USE_USART1
void ClearBuffer_USART1()
{
	tx_index_usart1 = 0;
	rx_index_usart1 = 0;
	uint16_t i=0;
	for (; i<MAX_CHARS; i++)
		buffer_usart1[i]=0;
}
#endif

void ClearBuffer_USART2()
{
	tx_index_usart2 = 0;
	rx_index_usart2 = 0;
	uint16_t i=0;
	for (; i<MAX_CHARS; i++)
		buffer_usart2[i]=0;
}

/******************************************************************************/


/*-----------------------------------------------------*/
/* Code below adapted from STMicroelectronics examples */
/*-----------------------------------------------------*/

/**
  * @brief  Inserts a delay time.
  * @param  nTime: specifies the delay time length, in 1 ms.
  * @retval None
  */
void Delay(__IO uint32_t nTime)
{
	TimingDelay = nTime;

	while(TimingDelay != 0);
}

/**
  * @brief  Decrements the TimingDelay variable.
  * @param  None
  * @retval None
  */
//void TimingDelay_Decrement(void)
void Decrement_Counters(void)
{
	if (TimingDelay != 0x00)
	{
		TimingDelay--;	// used for Delay() only
	}

	if (logTimer != 0x00)
	{
		logTimer--;		// used to keep track of log period
	}

	if (TimeOutSEN != 0x00)
	{
		TimeOutSEN--;	// separate timeout for sensor communication as SPI interface uses TimeOut during this period
	}

	if (TimeOut != 0x00)
	{
		TimeOut--;		// interface timeout
	}
	
	/* diskio_sd.c */
	disk_timerproc();	// FatFs timer requirements
}

#ifdef  USE_FULL_ASSERT

/**
  * @brief  Reports the name of the source file and the source line number
  *   where the assert_param error has occurred.
  * @param  file: pointer to the source file name
  * @param  line: assert_param error line source number
  * @retval None
  */
void assert_failed(uint8_t* file, uint32_t line)
{ 
  /* User can add his own implementation to report the file name and line number,
     ex: printf("Wrong parameters value: file %s on line %d\r\n", file, line) */

  /* Infinite loop */
	while (1)
	{}
}
#endif

/**
  * @}
  */

/* END OF FILE */
